import re

from setuptools import setup

setup(
    name="fastapis",
    version=re.search("(?smi)^__version__.+?([0-9.]+)", open("fastapis/__init__.py", encoding="utf-8").read())[1],
    packages=["fastapis"],
    install_requires=[
        "aiomysql==0.2.0",
        "aiosqlite==0.21.0",
        "annotated-types==0.7.0",
        "anyio==4.9.0",
        "asyncpg==0.30.0",
        "click==8.2.1",
        "colorama==0.4.6",
        "fastapi==0.115.12",
        "greenlet==3.2.2",
        "h11==0.16.0",
        "idna==3.10",
        "itsdangerous==2.2.0",
        "pydantic==2.11.5",
        "pydantic_core==2.33.2",
        "PyMySQL==1.1.1",
        "python-multipart==0.0.20",
        "PyYAML==6.0.2",
        "sniffio==1.3.1",
        "SQLAlchemy==2.0.41",
        "starlette==0.46.2",
        "typing-inspection==0.4.1",
        "typing_extensions==4.13.2",
        "uvicorn==0.34.2",
    ],
    long_description_content_type="text/markdown",
    url="https://github.com/ultrafunkamsterdam/fastapis",
    license="GPL-3.0",
    author="UltrafunkAmsterdam",
    author_email="info@blackhat-security.nl",
    description="SQLAlchemy with Fastapi asyncio for humans",

)
